include("covariance_functions.jl")
include("gaussian_ops.jl")

function bayesian_linear_regression_inf(x, y, weights_prior, lik_variance; print_debug=true)
	N = length(y)

	println("\nrunning Bayesian linear regression...")

	# compute posterior over w
	weights_posterior = Dict()
	weights_posterior["v"] = pinv(pinv(weights_prior["v"]) + lik_variance*(x'*x))
	weights_posterior["mu"] = weights_posterior["v"]*(weights_prior["v"] \ weights_prior["mu"]') + lik_variance*weights_posterior["v"]*x'*y

	# make predictions
	predictive_dist = Dict()
	predictive_dist["mu"] = x * weights_posterior["mu"]
	predictive_dist["v"] = zeros(N,1)
	for i=1:N
	    predictive_dist["v"][i] = (lik_variance + x[i,:] * weights_posterior["v"] * x[i,:]')[1]
	end

	posterior_lr = Dict("predictive_dist" => predictive_dist, "w" => weights_posterior)

	return posterior_lr
end

function bayesian_linear_regression_pred(x_test, posterior)
	return Dict("mu" => x_test * posterior["w"]["mu"])
end

function bayesian_linear_regression_marginal_lik(x, y, prior_variance, lik_variance; derivatives=false)
	return bayesian_linear_regression_marginal_lik(x, x*x', y, prior_variance, lik_variance; derivatives=derivatives)
end

function bayesian_linear_regression_marginal_lik(x, xx, y, prior_variance, lik_variance; derivatives=false)
	N = size(x,1)
	D = size(x,2)

	C = lik_variance*eye(N) + prior_variance*xx
	lZ = 0
	L = NaN; invC = NaN; invCy = NaN
	try
		L = chol(C)
		#invC = inv(C)
		invC = L\(L\eye(N))'
		invCy = invC*y
		lZ = -N*log(2*pi)/2 -sum(diag(L)) -y'*invCy/2
		#lZ = -N*log(2*pi)/2 -log(det(C))/2 -y'*invCy/2
	catch
        println("warning: inference method failed")
        if derivatives; return -Inf, [Inf,Inf]; else; return -Inf; end
	end

	if derivatives
		dlZ = zeros(2)
		invCxx = invC*xx
		dlZ[1] = -trace(invCxx)/2 + sum(y'*invCxx*invCy)/2
		dlZ[2] = -trace(invC)/2 + sum(y'*invC*invCy)/2

		return lZ[1,1], dlZ
	end

	return lZ[1,1]
end

function lr_optimize_hyperparams(x, y, prior_variance, lik_variance; max_iterations=100)
    global gX = x
    global gy = y
    global xx = x*x'
    global iter = 0
    D = size(gX,2)
    init_params = [prior_variance, lik_variance]
    #d = DifferentiableFunction(lr_f)
    d = DifferentiableFunction(lr_f, lr_g!, lr_fg!)
    println("optimizing hyper-parameters...")
    res = Optim.optimize(d, init_params, method=:cg, iterations=max_iterations)
    println()
    nlZ = res.f_minimum
    opt_params = res.minimum
    println("minimum lZ: ", -nlZ)
    println("num iterations: ", res.iterations)
    weights_prior = Dict("mu" => zeros(1,D), "v" => opt_params[1]*eye(D))
	lik_variance = opt_params[2]
    return weights_prior, lik_variance, -nlZ[1,1]
end

function lr_f(x::Vector)
    global gX, gy, xx, iter
    D = size(gX,2)
    iter += 1
	prior_variance = x[1]
	lik_variance = x[2]
    lZ = bayesian_linear_regression_marginal_lik(gX, xx, gy, prior_variance, lik_variance)
    #println("\r[eval $iter] lZ: $lZ       \t", x)
    print("\r[eval $iter] lZ: $lZ       \t")
    return -lZ
end

function lr_g!(x::Vector, storage::Vector)
    global gX, gy, xx, iter
    D = size(gX,2)
	prior_variance = x[1]
	lik_variance = x[2]
    lZ, dlZ = bayesian_linear_regression_marginal_lik(gX, xx, gy, prior_variance, lik_variance; derivatives=true)
    storage[:] = -dlZ[:]
end

function lr_fg!(x::Vector, storage::Vector)
    global gX, gy, xx, iter
    D = size(gX,2)
    iter += 1
	prior_variance = x[1]
	lik_variance = x[2]
    lZ, dlZ = bayesian_linear_regression_marginal_lik(gX, xx, gy, prior_variance, lik_variance; derivatives=true)
    #println("\r[eval $iter] lZ: $lZ       \t", x)
    print("\r[eval $iter] lZ: $lZ       \t")
    storage[:] = -dlZ[:]
    return -lZ
end

