include("covariance_functions.jl")
include("gaussian_ops.jl")
include("utils.jl")

function gaussian_gp_exact_inf(X, y, cov, sigma)
    N = size(X,1)
    D = size(X,2)
    sigma2 = exp(2*sigma)

    println("\nrunning Gaussian GP (exact)...")

    # compute covariance matrix
    K = covSEard(X, X, cov["v"], cov["l"])

    # exact bayesian solution
    p_f = Dict("mu" => zeros(N,1), "v" => zeros(N,1))
    print("inverting covariance matrix...")

    # compute inverse
    #K_sigma_inv = inv(K + sigma2*eye(N))

    # compute inverse, Rasmussen way (safer)
    L = chol(K/sigma2+eye(N))
    K_sigma_inv = solve_chol(L,eye(N))/sigma2

    println(" done!")
    print("estimiting f_i's...")
    for i=1:N
        p_f["mu"][i] = (K[i,:] * K_sigma_inv * y)[1]
        p_f["v"][i] = (K[i,i] - K[i,:] * K_sigma_inv * K[i,:]')[1]
    end
    println(" done!")

    posterior_gp = Dict("p_f" => p_f, "K_sigma_inv" => K_sigma_inv)

    return posterior_gp
end

function gaussian_gp_exact_pred(X, y, cov, sigma, X_test, posterior_gp)
    N_test = size(X_test,1)
    sigma2 = exp(2*sigma)
    predictive_dist_test = Dict("mu" => zeros(N_test), "v" => zeros(N_test))

    print("making predictions...")
    for i=1:N_test
        # compute k(X,x*)
        K_xstar = covSEard(X_test[i,:], X, cov["v"], cov["l"])
        # compute k(x*,x*)
        K_xstar_xstar = covSEard(X_test[i,:], X_test[i,:], cov["v"], cov["l"])

        # make prediction
        # <y*> = k(x*,X) (K + \sigma^2 I)^{-1} Y
        ystar_mean = (K_xstar * posterior_gp["K_sigma_inv"]) * y
        # compute prediction variance
        # var(y*)^2 = k(x*,x*) + \sigma^2 - k(X,x*).T (K + \sigma^2 I)^{-1} k(X,x*)
        ystar_var = K_xstar_xstar + sigma2 - (K_xstar * posterior_gp["K_sigma_inv"]) * K_xstar'

        predictive_dist_test["mu"][i] = ystar_mean[1,1]
        predictive_dist_test["v"][i] = ystar_var[1,1]
    end
    println(" done!")

    return predictive_dist_test
end

function gaussian_gp_exact_marginal_lik(X, y, cov, sigma; derivatives=false)
    N = size(X,1)
    D = size(X,2)
    sigma2 = exp(2*sigma)

    # compute covariance matrix
    K = covSEard(X, X, cov["v"], cov["l"])

    # compute marginal likelihood
    # 1) by the book
    #K_plus_sigma = K + sigma2*eye(N)
    #lZ = -0.5*y'*inv(K_plus_sigma)*y - 0.5*log(det(K_plus_sigma)) - (N/2)*log(2*pi)
    
    # 2) faster way
    #K_plus_sigma = K + sigma2*eye(N)
    #lZ = -0.5*y'*(K_plus_sigma\y) - 0.5*log(det(K_plus_sigma)) - (N/2)*log(2*pi)
    
    # 3) Rasmussen way with cholesky (safer way)
    L = nothing; alpha = nothing
    try
        L = chol(K/sigma2+eye(N))
        alpha = solve_chol(L,y)/sigma2 # alpha = (K+sigma)\y
    catch
        println("\nwarning: inference method failed")
        if derivatives; return [-Inf], Inf; else; return [-Inf]; end
    end
    sW = ones(N,1)/sqrt(sigma2)
    lZ = -y'*alpha/2 - sum(log(diag(L))) - N*log(2*pi*sigma2)/2

    if derivatives
        num_params = length(cov["l"]) + length(cov["v"]) + 1 # cov params + lik param
        dlZ = zeros(num_params)

        # covariance hyperparameters
        Q = -solve_chol(L,eye(N))/sigma2 + alpha*alpha'
        for p=1:num_params-1
            dK = covSEard(X, X, cov["v"], cov["l"], p=p)
            dlZ[p] = sum(Q.*dK)/2
        end

        # likelihood hyperparameters
        dlZ[end] = sigma2*trace(Q)

        return lZ[1,1], dlZ
    else
        return lZ[1,1]
    end
end
