function dampening(message, old_message, dampening_factor, iter)
    if dampening_factor < 1.0
        if iter > 1
            if haskey(message,"mu") && haskey(old_message,"mu")
                message["mu"] = dampening_factor * message["mu"] + (1-dampening_factor) * old_message["mu"]
                message["v"] = dampening_factor * message["v"] + (1-dampening_factor) * old_message["v"]
            end
            if haskey(message,"nu") && haskey(old_message,"nu")
                message["nu"] = dampening_factor * message["nu"] + (1-dampening_factor) * old_message["nu"]
                message["tau"] = dampening_factor * message["tau"] + (1-dampening_factor) * old_message["tau"]
            end
        end
        old_message = deepcopy(message)
    end
    return message, old_message
end

function momentToNaturalParameters!(p)
    p["nu"] = pinv(p["v"])
    #p["nu"] = cholfact(p["v"]) \ eye(size(p["v"],1))
    p["tau"] = p["nu"] * p["mu"]
end

function naturalToMomentParameters!(p)
    #p["v"] = pinv(p["nu"])
    p["v"] = p["nu"] \ eye(size(p["nu"],1))
    p["mu"] = p["v"] * p["tau"]
end

function gaussianProduct(gaussian1, gaussian2)
    gaussianResult = Dict{ASCIIString,Array{Float64,2}}()
    if haskey(gaussian1,"nu") && haskey(gaussian2,"nu") && haskey(gaussian1,"tau") && haskey(gaussian2,"tau")
        gaussianResult["nu"] = gaussian1["nu"] + gaussian2["nu"]
        gaussianResult["tau"] = gaussian1["tau"] + gaussian2["tau"]
    else
        gaussianResult["v"] = gaussian1["v"] - gaussian1["v"] * ((gaussian1["v"] + gaussian2["v"]) \ gaussian1["v"]) # Woodbury formula
        gaussianResult["mu"] = gaussianResult["v"] * ((gaussian1["v"]\gaussian1["mu"]) + (gaussian2["v"]\gaussian2["mu"]))
    end
    return gaussianResult
end

function gaussianDivision(gaussian1, gaussian2)
    gaussianResult = Dict{ASCIIString,Array{Float64,2}}()
    if haskey(gaussian1,"nu") && haskey(gaussian2,"nu") && haskey(gaussian1,"tau") && haskey(gaussian2,"tau")
        gaussianResult["nu"] = gaussian1["nu"] - gaussian2["nu"]
        gaussianResult["tau"] = gaussian1["tau"] - gaussian2["tau"]
    else
        gaussianResult["v"] = gaussian1["v"] - gaussian1["v"] * ((gaussian1["v"] - gaussian2["v"]) \ gaussian1["v"]) # Woodbury formula
        gaussianResult["mu"] = (gaussianResult["v"] * (pinv(gaussian1["v"])*gaussian1["mu"] - pinv(gaussian2["v"])*gaussian2["mu"]))
        #gaussianResult["mu"] = gaussianResult["v"] * ((gaussian1["v"]\gaussian1["mu"]) - (gaussian2["v"]\gaussian2["mu"]))
    end
    return gaussianResult
end

function gaussianMarginal(gaussian1, A, b, invL)
    gaussianResult = Dict{ASCIIString,Array{Float64,2}}()
    if haskey(gaussian1, "nu")
        #gaussianResult["v"] = A * (gaussian1["nu"] \ A') + invL
        gaussianResult["v"] = A * (gaussian1["nu"] \ A') + invL  # this version gives no problem with singular matrices
        #gaussianResult["mu"] = A * (gaussian1["nu"] \ gaussian1["tau"])
        gaussianResult["mu"] = A * (gaussian1["nu"] \ gaussian1["tau"])  # this version gives no problem with singular matrices
    else
        gaussianResult["v"] = A * gaussian1["v"] * A' + invL
        gaussianResult["mu"] = A * gaussian1["mu"] + b
    end
    return gaussianResult
end

function gaussianTruncation(gaussian1, l)
    gaussianResult = Dict{ASCIIString,Array{Float64,2}}()
    if !haskey(gaussian1,"mu")
        error("not implemented yet")
    else
        if any(gaussian1["v"] .>= 0.0)
            sqrt_v = sqrt(gaussian1["v"])
        else
            sqrt_v = - sqrt(abs(gaussian1["v"]))
        end
        eta = (gaussian1["mu"] - l) / sqrt_v
        pdf_over_cdf, Z = gauOverCumGauss(eta)
        pdf_over_cdf = pdf_over_cdf[1]
        gaussianResult["mu"] = gaussian1["mu"] + sqrt_v * pdf_over_cdf
        gaussianResult["v"] = gaussian1["v"] * (1 - eta * pdf_over_cdf - pdf_over_cdf.^2)
    end
    return gaussianResult, Z
end

function gauOverCumGauss(f)
    if size(f) == ()
        f = [f]
    end
    p = (1+erf(f/sqrt(2.0)))/2.0

    n_p = zeros(size(f))       # safely compute Gaussian over cumulative Gaussian
    ok = f.>-5                            # naive evaluation for large values of f
    n_p[ok] = (exp(-f[ok].^2/2)/sqrt(2*pi)) ./ p[ok]

    bd = f.<-6                                      # tight upper bound evaluation
    n_p[bd] = sqrt(f[bd].^2/4+1)-f[bd]/2

    interp = ~ok & ~bd                # linearly interpolate between both of them
    tmp = f[interp]
    lam = -5-f[interp]
    n_p[interp] = (1-lam).*(exp(-tmp.^2/2)/sqrt(2*pi))./p[interp] + lam .*(sqrt(tmp.^2/4+1)-tmp/2)

    return n_p, p # return (pdf/cdf) and cdf
end

function gaussianCDF(z)
    return (1+erf(z/sqrt(2.0)))/2.0
end

