include("covariance_functions.jl")
include("gaussian_ops.jl")
include("gauher.jl")
include("utils.jl")

function laplace_moments(y, mu)
    lg = mu
    dlg = ones(size(mu,1))
    d2lg = zeros(size(mu,1))
    d3lg = zeros(size(mu,1))
    elg = exp(lg)
    lpi = lg.*y - elg - lgamma(y+1)
    dlpi = dlg * (y-elg)
    d2lpi = d2lg.*(y-elg) - dlg.*dlg.*elg
    return lpi, dlpi, d2lpi
end

function lik_ep_quad(y, mu, s2)
    N_quad = 20
    n = size(y,1)
    t,w = gauher(N_quad)
    lw = log(w')
    y = y*ones(n,1)
    mu = mu*ones(n,1)
    sig = sqrt(s2)*ones(n,1)

    # Laplace approximation
    lpi, dlpi, d2lpi = laplace_moments(y*ones(1,N_quad), sig*t'+mu*ones(1,N_quad))
    lZ,_ = logsumexp(lpi+lw)
    a = exp(lpi - lZ*ones(1,N_quad) + lw)
    dlZ = sum(a.*dlpi,2)
    d2lZ = sum(a.*(d2lpi+dlpi.*dlpi),2) - dlZ.*dlZ
    return lZ, dlZ, d2lZ
end

function poisson_gp_ep_inf(X, y, cov; verbose=true)
    N = size(X,1)
    D = size(X,2)

    if verbose
        println("\nrunning Poisson GP (EP)...")
        println("iter 0: initialization")
    end

    # compute covariance matrix
    K = covSEard(X, X, cov["v"], cov["l"])

    # data structure for storing the approximate posterior on f
    q_f = Dict("mu" => zeros(N,1), "v" => K)

    # data structures for storing the approximate f_i marginals
    q_fi = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # messages from f_i variables to g factor
    m_fi_g = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fi_g_old = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # messages from g factor to f_i variables
    m_g_fi = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_g_fi_old = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # initialize messages and marginals approximations
    for i=1:N
        q_fi[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_fi_g[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_fi_g_old[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_g_fi[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_g_fi_old[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
    end

    # run EP
    max_iters = 15
    dampening_factor = 0.8
    tsigma = zeros(N,1)     # tilde parameters 
    tnu = zeros(N,1)   
    ttau = zeros(N,1)
    nu_ni = zeros(N,1)      # cavity parameters
    tau_ni = zeros(N,1)
    yi_marginals = zeros(N,1)
    Zi_hat = zeros(N,1)
    L = NaN
    for iter=1:max_iters
        print("iter $iter: ")
        
        yi_marginals = zeros(N,1)
        
        for i=1:N # sweep through the data points
            #println("processing data point $i, y($i)=$(y[i])")
            
            # step 1: compute the cavity distribution
            m_g_fi[i] = gaussianDivision(Dict("nu" => ones(1,1)*(q_f["mu"][i]/q_f["v"][i,i]), "tau" => ones(1,1)*(1/q_f["v"][i,i])), m_fi_g[i])
            m_g_fi[i], m_g_fi_old[i] = dampening(m_g_fi[i], m_g_fi_old[i], dampening_factor, iter)
            tau_ni[i] = m_g_fi[i]["tau"][1,1]
            nu_ni[i] = m_g_fi[i]["nu"][1,1]

            # step 2: compute approx. posterior on f_i using Gauss-Hermite quadrature
            lZ, dlZ, d2lZ = lik_ep_quad(y[i], m_g_fi[i]["nu"]/m_g_fi[i]["tau"], 1./m_g_fi[i]["tau"])
            q_fi[i]["mu"] = -lZ
            yi_marginals[i] = q_fi[i]["mu"][1,1]
            
            # step 3: compute the new approximating factor \tilde{f}_i that has the desired moments
            ttau_val = -d2lZ / (1+d2lZ/tau_ni[i])
            ttau_val = max(ttau_val,0.1) # enforce positivity i.e. lower bound ttau by zero
            tnu_val = ( dlZ - nu_ni[i]/tau_ni[i]*d2lZ )/(1+d2lZ/tau_ni[i])
            m_fi_g[i] = Dict("nu" => ones(1,1)*tnu_val, "tau" => ones(1,1)*ttau_val)
            m_fi_g[i], m_fi_g_old[i] = dampening(m_fi_g[i], m_fi_g_old[i], dampening_factor, iter)
            tsigma[i] = 1/m_fi_g[i]["tau"][1,1]
            tnu[i] = m_fi_g[i]["nu"][1,1]
            ttau[i] = m_fi_g[i]["tau"][1,1]
        end

        # step 0: compute new posterior over f
        L = chol(diagm(tsigma[:]) + K)
        q_f["v"] = K - K * solve_chol(L,K)
        q_f["mu"] = q_f["v"] * tnu

        if verbose
            print("y: ")
            cc, mae, rmse, rae, rrse = compute_error_stats(y, yi_marginals)
        end
    end

    posterior_gp = Dict("q_f"=>q_f, "K"=>K, "tnu"=>tnu, "ttau"=>ttau, "tsigma"=>tsigma, "nu_ni"=>nu_ni, "tau_ni"=>tau_ni, "Zi_hat"=>Zi_hat, "L"=>L)

    return posterior_gp
end

function poisson_gp_ep_pred(X, y, cov, X_test, posterior_gp)
    N_test = size(X_test,1)

    # compute inverse
    #K_tsigma_inv = inv(posterior_gp["K"] + posterior_gp["tsigma"])

    # compute inverse, Rasmussen way (safer)
    K_tsigma_inv = solve_chol(posterior_gp["L"],eye(length(y)))

    # compute predictions
    predictive_dist_test = Dict("mu" => zeros(N_test), "v" => zeros(N_test))
    print("making predictions...")
    N_quad = 20
    t,w = gauher(N_quad)
    lw = ones(1,1)*log(w')
    for i=1:N_test
        # compute k(X,x*)
        K_xstar = covSEard(X_test[i,:], X, cov["v"], cov["l"])
        # compute k(x*,x*)
        K_xstar_xstar = covSEard(X_test[i,:], X_test[i,:], cov["v"], cov["l"])

        # estimate posterior value
        ystar_mean = (K_xstar * K_tsigma_inv) * posterior_gp["q_f"]["mu"]
        ystar_var = K_xstar_xstar - (K_xstar * K_tsigma_inv) * K_xstar'
        #push!(predictive_dist_test, ystar_mean[1,1])

        # compute predictive posterior
        # first moment using Gaussian-Hermite quad
        mu = ystar_mean.*ones(1,1)
        sig = sqrt(ystar_var).*ones(1,1)
        lg = sig*t'+mu*ones(1,N_quad)
        lZ,_ = logsumexp(lg+lw)
        ymu = exp(lZ)
        # second y moment from Poisson distribution
        elg = exp(lg)
        yv = elg
        ys2 = (yv+(elg-ymu*ones(1,N_quad)).^2)*w

        predictive_dist_test["mu"][i] = ymu[1,1]
        predictive_dist_test["v"][i] = ys2[1,1]
    end
    println(" done!")

    return predictive_dist_test
end

