include("covariance_functions.jl")
include("gaussian_ops.jl")
include("utils.jl")

function truncated_gp_ep_inf(X, y, cov, sigma; verbose=true)
    N = size(X,1)
    D = size(X,2)
    sigma2 = exp(2*sigma)

    if verbose
        println("\nrunning Truncated GP (EP)...")
        println("iter 0: initialization")
    end

    # compute covariance matrix
    K = covSEard(X, X, cov["v"], cov["l"])

    # data structure for storing the approximate posterior on f
    q_f = Dict("mu" => zeros(N,1), "v" => K)

    # data structures for storing the approximate f_i marginals
    q_fi = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # messages from f_i variables to g factor
    m_fi_g = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_fi_g_old = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # messages from g factor to f_i variables
    m_g_fi = Array(Dict{ASCIIString,Array{Float64,2}},N)
    m_g_fi_old = Array(Dict{ASCIIString,Array{Float64,2}},N)

    # initialize messages and marginals approximations
    for i=1:N
        q_fi[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_fi_g[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_fi_g_old[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_g_fi[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
        m_g_fi_old[i] = Dict("nu" => zeros(1,1), "tau" => zeros(1,1))
    end

    # run EP
    max_iters = 5
    dampening_factor = 1.0
    tsigma = zeros(N,1)     # tilde parameters 
    tnu = zeros(N,1)   
    ttau = zeros(N,1)
    nu_ni = zeros(N,1)      # cavity parameters
    tau_ni = zeros(N,1)
    yi_marginals = zeros(N,1)
    Zi_hat = zeros(N,1)
    L = NaN
    for iter=1:max_iters
        if verbose; print("iter $iter: "); end
        
        for i=1:N # sweep through the data points
            #println("processing data point $i, y($i)=$(y[i])")
            
            # step 1: compute the cavity distribution
            m_g_fi[i] = gaussianDivision(Dict("nu" => ones(1,1)*(q_f["mu"][i]/q_f["v"][i,i]), "tau" => ones(1,1)*(1/q_f["v"][i,i])), m_fi_g[i])
            m_g_fi[i], m_g_fi_old[i] = dampening(m_g_fi[i], m_g_fi_old[i], dampening_factor, iter)
            tau_ni[i] = m_g_fi[i]["tau"][1,1]
            nu_ni[i] = m_g_fi[i]["nu"][1,1]
            
            # step 2: compute approximate posterior on f_i (site parameters)
            p_hat = gaussianProduct(m_g_fi[i], Dict("nu" => y[i]/sigma2, "tau" => 1.0/sigma2))
            p_hat["v"] = 1./p_hat["tau"]
            p_hat["mu"] = p_hat["nu"] * p_hat["v"]
            trunc_dist, Zi = gaussianTruncation(p_hat, 0)
            q_fi[i]["tau"] = 1./trunc_dist["v"]
            q_fi[i]["nu"] = q_fi[i]["tau"] * trunc_dist["mu"]
            Zi_hat[i] = Zi[1,1]
            yi_marginals[i] = trunc_dist["mu"][1]
            
            # step 3: compute the new approximating factor \tilde{f}_i that has the desired moments
            m_fi_g[i] = gaussianDivision(q_fi[i], m_g_fi[i])
            m_fi_g[i], m_fi_g_old[i] = dampening(m_fi_g[i], m_fi_g_old[i], dampening_factor, iter)
            tsigma[i] = 1/m_fi_g[i]["tau"][1,1]
            tnu[i] = m_fi_g[i]["nu"][1,1]
            ttau[i] = m_fi_g[i]["tau"][1,1]
        end
        
        # step 0: compute new posterior over f
        L = chol(diagm(tsigma[:]) + K)
        q_f["v"] = K - K * solve_chol(L,K)
        q_f["mu"] = q_f["v"] * tnu

        if verbose
            print("y: ")
            cc, mae, rmse, rae, rrse = compute_error_stats(y, q_f)
        end
    end
    posterior_gp = Dict("q_f"=>q_f, "K"=>K, "tnu"=>tnu, "ttau"=>ttau, "tsigma"=>tsigma, "nu_ni"=>nu_ni, "tau_ni"=>tau_ni, "Zi_hat"=>Zi_hat, "L"=>L)

    return posterior_gp
end

function truncated_gp_ep_pred(X, y, cov, sigma, X_test, posterior_gp)
    N_test = size(X_test,1)
    sigma2 = exp(2*sigma)

    # compute inverse
    #K_tsigma_inv = inv(posterior_gp["K"] + posterior_gp["tsigma"])

    # compute inverse, Rasmussen way (safer)
    K_tsigma_inv = solve_chol(posterior_gp["L"],eye(length(y)))

    # compute predictions
    predictive_dist_test = Dict("mu" => zeros(N_test), "v" => zeros(N_test))
    print("making predictions...")
    for i=1:N_test
        # compute k(X,x*)
        K_xstar = covSEard(X_test[i,:], X, cov["v"], cov["l"])
        # compute k(x*,x*)
        K_xstar_xstar = covSEard(X_test[i,:], X_test[i,:], cov["v"], cov["l"])

        # estimate f_i posterior value
        ystar_mean = (K_xstar * K_tsigma_inv) * y
        ystar_var = K_xstar_xstar + sigma2 - (K_xstar * K_tsigma_inv) * K_xstar'

        # compute predictive posterior
        pred_post, Z = gaussianTruncation(Dict("mu" => ystar_mean, "v" => ystar_var), 0)

        predictive_dist_test["mu"][i] = pred_post["mu"][1,1]
        predictive_dist_test["v"][i] = pred_post["v"][1,1]
    end
    println(" done!")

    return predictive_dist_test
end

function truncated_gp_ep_marginal_lik(X, y, cov, sigma; derivatives=false)
    N = size(X,1)
    sigma2 = exp(2*sigma)

    post = Dict()
    try
        post = truncated_gp_ep_inf(X, y, cov, sigma; verbose=false)
    catch
        println("\nwarning: inference method failed")
        if derivatives; return [-Inf], Inf; else; return [-Inf]; end
    end
    sW = sqrt(post["ttau"][:])
    L = chol(eye(N)+sW*sW'.*post["K"])
    v = diag(post["q_f"]["v"])
    lZ = -sum(log(diag(L))) + post["tnu"]'*post["q_f"]["v"]*post["tnu"]/2 - (v'*post["tnu"].^2)/2
    lZ += post["nu_ni"]'*((post["ttau"]./post["tau_ni"].*post["nu_ni"] - 2*post["tnu"]).*v)/2 
    lZ += sum(log(1+post["ttau"]./post["tau_ni"]))/2
    mu_ni = post["nu_ni"]./post["tau_ni"]
    v_ni = 1.0./post["tau_ni"]
    for i=1:N
        lZi = -(y[i]-mu_ni[i]).^2./(sigma2+v_ni[i])/2 - log(2*pi*(sigma2+v_ni[i]))/2
        lZi += log(post["Zi_hat"][i])
        lZ += lZi
    end

    if derivatives
        num_params = length(cov["l"]) + length(cov["v"]) + 1 # cov params + lik param
        dlZ = zeros(num_params)

        # covariance hyperparameters
        alpha = post["tnu"] - sW.*solve_chol(L,sW.*(post["K"]*post["tnu"]))
        F = alpha*alpha' - repmat(sW,1,N) .* solve_chol(L,diagm(sW))
        for p=1:num_params-1
            dK = covSEard(X, X, cov["v"], cov["l"], p=p)
            dlZ[p] = sum(F.*dK)/2
        end

        # likelihood hyperparameters
        dlZhyp = ((y - mu_ni).^2./(sigma2 + v_ni)-1) ./ (1+v_ni./sigma2)  # deriv. w.r.t. likelihood hyper-parameter
        #dlZhyp = 0
        dlZhyp = sum(dlZhyp)
        for i=1:N
            p_hat = gaussianProduct(Dict("mu" => ones(1,1)*mu_ni[i], "v" => ones(1,1)*v_ni[i]), Dict("mu" => y[i], "v" => sigma2))
            zi = p_hat["mu"][1,1]/sqrt(p_hat["v"][1,1])
            l_p, p = gauOverCumGauss(zi)
            dzi = -y[i] / (sigma2*sigma2*sqrt(p_hat["v"][1,1]))
            dzi += (0.5 * zi) / (sigma2*sigma2*sqrt(p_hat["v"][1,1]))
            dlZhyp += l_p[1,1] * dzi
        end
        dlZ[end] = dlZhyp

        return lZ[1,1], dlZ
    else
        return lZ[1,1]
    end
end

