# solve_chol - solve linear equations from the Cholesky factorization.
# Solve A*X = B for X, where A is square, symmetric, positive definite. The
# input to the function is R the Cholesky decomposition of A and the matrix B.
# Example: X = solve_chol(chol(A),B);
function solve_chol(L, B)
    return L\(L'\B)
end

# computes y = log( sum(exp(x),2) ), the softmax in a numerically safe way by
# subtracting the row maximum to avoid cancelation after taking the exp
# the sum is done along the rows
function logsumexp(logx)
  N = size(logx,2)
  max_logx = maximum(logx)
  # we have all values in the log domain, and want to calculate a sum
  x = exp(logx-max_logx)
  y = log(sum(x,2)) + max_logx
  return y,x
end

function compute_error_stats(trues, preds; print=true)
	if isa(preds,Dict)
	    preds = preds["mu"]
	end
	if length(preds) == 0
		@printf("cc=NA\tmae=NA\trmse=NA\trae=NA\trrse=NA\tr2=NA\n")
		return 0,0,0,0,0,0
	end

	@assert length(trues)==length(preds)

	cc = cor(preds, trues)[1]
	mae = sum(abs(trues - preds)) / length(trues)
	rmse = sqrt(sum((trues - preds).^2) / length(trues))
	rae = 100*sum(abs(trues - preds)) / sum(abs(trues - mean(trues)))
	rrse = 100*sqrt(sum((trues - preds).^2) / sum((trues - mean(trues)).^2))
	r2 = max(0, 1 - sum((trues-preds).^2) / sum((trues-mean(trues)).^2))

	if print
		@printf("cc=%.3f\tmae=%.3f\trmse=%.3f\trae=%.3f\trrse=%.3f\tr2=%.3f\n", cc, mae, rmse, rae, rrse, r2)
	end
	return cc, mae, rmse, rae, rrse, r2
end
